import json
import math
import logging
from chirptext import DataObject as DObj
from chirptext import chio
from chirptext.leutile import hamilton_allocate
from .colors import to_color_list, FIXED_LANGUAGES


def getLogger():
    return logging.getLogger(__name__)


class SkillMatrix:
    def __init__(self, understanding, reading, writing, speaking, max_value=10):
        self.understanding = understanding
        self.reading = reading
        self.writing = writing
        self.speaking = speaking
        self.max_value = max_value

    @property
    def max_value(self):
        return self.__max_value

    @max_value.setter
    def max_value(self, value):
        if value is not None and value <= 0:
            raise Exception("Max value of a skill matrix must be positive")
        self.__max_value = value
        
    @property
    def top(self):
        return self.understanding

    @property
    def right(self):
        return self.reading

    @property
    def bottom(self):
        return self.writing
    
    @property
    def left(self):
        return self.speaking

    def to_tuple(self):
        return (self.understanding, self.reading, self.writing, self.speaking)

    def __iter__(self):
        return iter(self.to_tuple())

    def __repr__(self):
        return str((self.understanding, self.reading, self.writing, self.speaking))

    def is_valid(self, strict=True):
        if strict:
            MIN = 0
        else:
            MIN = -1  # allow missing values
        if self.understanding >= MIN and self.reading >= MIN and self.writing >= MIN and self.speaking >= MIN:
            if self.max_value:
                return self.understanding <= self.max_value and self.reading <= self.max_value and self.writing <= self.max_value and self.speaking <= self.max_value
            else:
                return True
        return False

    def __str__(self):
        return repr(self)


class LEOPerson(DObj):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.languages = dict()

    def validate(self, messages=None, require_matrix=True, strict=True, hamilton=False):
        if messages is None:
            messages = []
        dist_sum = 0
        _myname = self.name.upper() if self.name else '<<UNKNOWN PERSON>>'
        if not self.name:
            messages.append("There is a person without name")
        for langname, lang in self.languages.items():
            if lang.dist == -1 and strict:
                messages.append(f"lang dist score of {_myname} - language: {langname} is negative")
            elif lang.dist > 100:
                messages.append(f"lang dist score of {_myname} - language: {langname} is {lang.dist} (> 100)")
            dist_sum += lang.dist
            if require_matrix and not lang.skills.is_valid(strict=strict):
                lang_tuple = lang.skills.to_tuple()
                if lang_tuple == (-1, -1, -1, -1):
                    messages.append(f"{_myname} does not have a spider chart")
                else:
                    messages.append(f"SkillMatrix of {_myname} contains invalid value(s) -> {lang_tuple}")
        if not len(self.languages):
            messages.append(f"{_myname} does not have language info (language name, scores, etc.)")
        if not math.isclose(dist_sum, 100.0):
            lang_dists = [str(lang.dist) for lang in self.languages.values()]
            if hamilton:
                getLogger().warning(f"  -- Language skill of {_myname} does not sum up to 100%  |  {' + '.join(lang_dists)} = {dist_sum}")
                lang_keys = self.languages.keys()
                old_values = [self.languages[k].dist for k in lang_keys]
                new_values = hamilton_allocate(old_values)
                for k, v in zip(lang_keys, new_values):
                    self.languages[k].dist = v
                getLogger().warning(f"  -- Fixed using Hamilton method to: {lang_keys} | {new_values}")
            else:
                messages.append(f"Language skill of {_myname} does not sum up to 100%  |  {' + '.join(lang_dists)} = {dist_sum}")
        return messages

    def to_dict(self):
        a_dict = super().to_dict()
        language_matrix = []
        for k, v in self.languages.items():
            language_matrix.append((k, v.dist, *v.skills))
        a_dict['languages'] = language_matrix
        return a_dict

    @staticmethod
    def from_dict(a_dict):
        person = LEOPerson(**a_dict)
        person.languages.update({x[0]: LanguageSkill(x[1], x[2:]) for x in a_dict['languages']})
        return person


class LanguageSkill(DObj):
    def __init__(self, dist, skills, **kwargs):
        """ A person's language skills (understanding, reading, writing, speaking)  """
        super().__init__(dist=dist, skills=SkillMatrix(*skills), **kwargs)


class LEOProfile(DObj):
    def __init__(self, reportID, *args, languages=None, **kwargs):
        """
        LEO Profile Data Structure
        """
        super().__init__(reportID=reportID, **kwargs)
        if languages is None:
            self.languages = []
        else:
            self.languages = [l for l in languages if l not in FIXED_LANGUAGES]

    def output_filename(self):
        if self.pdf_filename:
            return self.pdf_filename
        elif self.reportID:
            return f"{self.reportID}.pdf"
        else:
            return "report.pdf"

    def to_dict(self):
        a_dict = super().to_dict()
        a_dict['baby'] = self.baby.to_dict()
        a_dict['people'] = [p.to_dict() for p in self.people]
        return a_dict

    def to_json(self, **kwargs):
        return json.dumps(self.to_dict(), **kwargs)

    def gather_languages(self):
        languages = {k for k, v in self.baby.languages.items()}
        languages.update(self.languages)
        for person in self.people:
            languages.update(k for k, v in person.languages.items())
        return languages

    def get_extra_language_legends(self, limit=4):
        return [l for l in self.gather_languages() if l not in FIXED_LANGUAGES][:limit]

    def build_language_color_map(self):
        languages = self.gather_languages()
        colors = to_color_list(languages)
        return {l: c for l, c in zip(languages, colors)}

    def validate(self, strict=True, require_matrix=True, hamilton=False):
        messages = []
        if not self.reportID or not str(self.reportID).strip():
            messages.append("ReportID cannot be empty")
        if not self.baby.name:
            messages.append("Baby name is empty")
        if not self.baby.age:
            messages.append("Baby age is empty")
        if not self.baby.languages:
            messages.append("Baby language info is missing")
        else:
            self.baby.validate(messages, require_matrix=False, strict=strict, hamilton=hamilton)
        for person in self.people:
            person.validate(messages, require_matrix=require_matrix, strict=strict, hamilton=hamilton)
        return messages

    @staticmethod
    def from_dict(a_dict):
        reportID = ''
        kwargs = {}
        for k, v in a_dict.items():
            if k == 'reportID':
                reportID = v
            elif k == 'cert_date':
                kwargs['cert_date'] = v
            elif k == 'pdf_filename':
                kwargs[k] = v
            else:
                if k == 'baby':
                    v = LEOPerson.from_dict(v)
                elif k == 'people':
                    v = [LEOPerson.from_dict(p) for p in v]
                kwargs[k] = v
        return LEOProfile(reportID, **kwargs)

    @staticmethod
    def from_json_str(json_string):
        return LEOProfile.from_dict(json.loads(json_string))

    @staticmethod
    def from_file(input_path):
        return LEOProfile.from_json_str(chio.read_file(input_path))


def read_json(input_path):
    return LEOProfile.from_file(input_path)



FIXED_LANGUAGES = {"English", "Tamil", "Malay", "Mandarin"}
CN_DIALECTS = ['Cantonese', 'Hainanese', 'Hakka', 'Henghua', 'Hokchew',
               'Hokchia', 'Hokkien', 'Jiangxi', 'Shanghainese', 'Sichuan', 'Teochew']
BAHASA_DIALECTS = ['Bahasa Indonesia', 'Javanese', 'Boyanese', 'Malaysian Dialect', 'Other Bahasa']
INDIAN_DIALECTS = ['Bengali', 'Gujarati', 'Hindi', 'Punjabi', 'Urdu', 'Malayalam', "Telugu"]
CN_DIALECT_COLORS = ['#FF7518', '#E69F00', '#FF9966', '#FFCC99']
BAHASA_DIALECT_COLORS = ['#336633', '#009933']
INDIAN_DIALECT_COLORS = ["#663399", "#702963"]
OTHER_COLORS = ['#9B111E', '#999999', '#996515', '#000080', '#FF0000', '#6F4E37', '#636C77', '#C21E56', '#1C39BB', '#464647', '#5B342E']
COLOR_MAP = {
    'English': '#0072B2',
    'Mandarin': '#D55E00',
    'Malay': '#009E73',
    'Tamil': '#CC79A7',
    'Arabic': '#F0E442'
}
COLORS = ['#0072B2', '#D55E00', '#009E73', '#CC79A7', '#E69F00',
          '#FF9966', '#FFCC99', '#336633', '#009933', '#663399',
          '#53239A', '#F0E442', '#999999']


def map_language(language, prev_languages, color_options, reuse_lastcolor=False):
    if len(prev_languages) < len(color_options):
        color = color_options[len(prev_languages)]
        prev_languages.append(language)
    elif reuse_lastcolor:
        return color_options[-1]
    else:
        color = None
    return color


def to_color_list(languages, reuse_lastcolor=False):
    color_list = []
    _bahasa_dialects = []
    _cn_dialects = []
    _indian_dialects = []
    _others = []
    for language in (x.strip().title() for x in languages):
        color = None
        if language in COLOR_MAP:
            color = COLOR_MAP[language]
        elif language in CN_DIALECTS:
            color = map_language(language, _cn_dialects, CN_DIALECT_COLORS)
        elif language in BAHASA_DIALECTS:
            color = map_language(language, _bahasa_dialects, BAHASA_DIALECT_COLORS)
        elif language in INDIAN_DIALECTS:
            color = map_language(language, _indian_dialects, INDIAN_DIALECT_COLORS)
        if color is None:
            color = map_language(language, _others, OTHER_COLORS, reuse_lastcolor=reuse_lastcolor)
        if color is None:
            raise Exception("There are not enough colours to choose from")
        color_list.append(color)
    return color_list



import os
import logging
from pathlib import Path

from pyinkscape import Canvas, PieChart
from pyinkscape.charts import STYLE_REDDOT
from pyinkscape.styles import Style
from pyinkscape.render import prepare_output_dir, svg_to_pdf, merge_pdf
from .charts import SpiderChart


# ------------------------------------------------------------------------------
# Functions
# ------------------------------------------------------------------------------

def getLogger():
    return logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# Styles
# ------------------------------------------------------------------------------
TEMPLATE_DIR = Path(os.path.dirname(__file__)) / 'templates'
PAGE1_TEMPLATE = TEMPLATE_DIR / 'leo_page1.svg'
PAGE2_TEMPLATE = TEMPLATE_DIR / 'leo_page2.svg'
PAGE3_TEMPLATE = TEMPLATE_DIR / 'leo_page3.svg'
PIE_RADIUS = 57
CHART_RADIUS = (57, 57)
PEOPLE_NAME_STYLE = Style(font_size='16px', font_family='D-DINCondensed', font_weight="Bold", fill='#000000', fill_opacity='1', text_anchor="middle", dominant_baseline="hanging")

# ----------------------------------------
# Page 1
P1_BABY_NAME_LOC = (297, 75.69)
P1_BABY_AGE_LOC = (297, 122.1)
P1_BABY_NAME_STYLE = Style(font_size="50px", fill="#232323", font_style="normal", font_variant="normal",
                           font_weight="bold", font_stretch="normal", font_family='D-DIN Condensed',
                           text_anchor="middle", dominant_baseline="central")
P1_BABY_AGE_STYLE = Style(font_size="15px", fill="#000000", font_style="normal", font_variant="normal",
                          font_weight="bold", font_stretch="normal", font_family='D-DIN Condensed',
                          text_anchor="middle", dominant_baseline="central")
# ----------------------------------------
# Page 2
P2_BABY_NAME_LOC = (30, 82)
P2_BABY_AGE_LOC = (31, 120)
P2_BABY_NAME_STYLE = Style(font_size="45px", fill="#232323", font_style="normal", font_variant="normal",
                           font_weight="bold", font_stretch="normal", font_family='D-DIN Condensed',
                           text_anchor="start", dominant_baseline="central")
P2_BABY_AGE_STYLE = Style(font_size="14px", fill="#000000", font_style="normal", font_variant="normal",
                          font_weight="bold", font_stretch="normal", font_family='D-DIN Condensed',
                          text_anchor="start", dominant_baseline="central")
P2_BABY_PIE_LOC = (495, 118)
P2_PIE_COLS = (104, 241, 379, 517)
P2_PIE_LINE = 323
P2_NAME_LINE = 390
P2_USAGE_LINE = 412
P2_SPIDER_LINE = 500

# ----------------------------------------
# Page 3
P3_LANG_NAME_STYLE = Style(font_size='15px', font_family='D-DINCondensed', font_weight="Bold", fill='#000000', fill_opacity='1', text_anchor="start", dominant_baseline="central")
P3_PEOPLE_NAME_STYLE = Style(font_size='16px', font_family='D-DINCondensed', font_weight="Bold", fill='#000000', fill_opacity='1', text_anchor="middle", dominant_baseline="hanging")
# language legends
P3_LANG_ROW1 = 139.27
P3_LANG_ROW2 = 170
P3_LANG_BALL_COLS = (35, 182, 326, 471)
P3_LANG_TEXT_COLS = (54, 201, 343, 491)
# pie chart & names
P3_PIE_ROW = 323.27
P3_PIE_COLS = (82, 219, 360, 500)
# names & usage lines
P3_NAME_ROW = 390
P3_USAGE_ROW = 412
# spider chart
P3_SPIDER_ROW = 500


def generate_leo(profile, output_dir='output', overwrite=False, auto_clean=True, mkdir=False):
    ''' Generate LEO report '''
    output_dir = prepare_output_dir(output_dir, mkdir=mkdir)
    # generate color map
    color_map = profile.build_language_color_map()
    # page 1
    getLogger().info("Generating page 1 SVG")
    t1 = Canvas(PAGE1_TEMPLATE)
    g1 = t1.group('leo')
    g1.text(profile.baby.name.upper(), center=P1_BABY_NAME_LOC, style=P1_BABY_NAME_STYLE, id_prefix="p1_BabyName")
    g1.text(profile.baby.age.upper(), center=P1_BABY_AGE_LOC, style=P1_BABY_AGE_STYLE, id_prefix="p1_BabyAge")
    # page 2
    getLogger().info("Generating page 2 SVG")
    t2 = Canvas(PAGE2_TEMPLATE)
    g2 = t2.group('leo')
    # name & age
    g2.text(profile.baby.name.upper(), center=P2_BABY_NAME_LOC, style=P2_BABY_NAME_STYLE, id_prefix="p2_BabyName")
    g2.text(profile.baby.age.upper(), center=P2_BABY_AGE_LOC, style=P2_BABY_AGE_STYLE, id_prefix="p2_BabyAge")
    # baby's chart
    baby_languages = list(profile.baby.languages.keys())
    baby_colors = [color_map[l] for l in baby_languages]
    baby_slides = (x.dist for lang, x in profile.baby.languages.items())
    pie = PieChart(g2, center=P2_BABY_PIE_LOC, radius=CHART_RADIUS, colors=baby_colors)
    pie.slide(*baby_slides)
    pie.render(id_prefix="p2_baby_piechart")
    for i, person in enumerate(profile.people[:4]):
        languages = [lang for lang, p in person.languages.items() if p.dist > 0]
        colors = [color_map[l] for l in languages]
        g2.text(person.name.strip().upper(), center=(P2_PIE_COLS[i], P2_NAME_LINE), style=PEOPLE_NAME_STYLE, id_prefix=f"p2_person{i}_name")
        if person.usage != -1:
            g2.text(f"{round(person.usage)}%", center=(P2_PIE_COLS[i], P2_USAGE_LINE), style=PEOPLE_NAME_STYLE, id_prefix=f"p2_person{i}_usage")
        else:
            g2.text(f"- N/A -", center=(P2_PIE_COLS[i], P2_USAGE_LINE), style=PEOPLE_NAME_STYLE, id_prefix=f"p2_person{i}_usage")
        # render pie charts
        slides = [x.dist for x in person.languages.values() if x.dist > 0]
        if not person.piesize:
            _adjust = 100
        elif person.piesize > 10:
            _adjust = person.piesize
        else:
            # draw draw the pie chart too small
            _adjust = 10
        radius = (PIE_RADIUS * _adjust / 100, PIE_RADIUS * _adjust / 100)
        getLogger().debug(f"{person.name} -- adjust -> {_adjust}")
        pie = PieChart(g2, center=(P2_PIE_COLS[i], P2_PIE_LINE), radius=radius, colors=colors)
        pie.slide(*slides)
        pie.render(id_prefix=f"p2_person{i}_piechart")
        # render spider charts
        spider_matrix = [x.skills for x in person.languages.values() if x.skills.is_valid()]
        if not spider_matrix:
            getLogger().warning(f": {profile.reportID}/page 2 - person `{person.name.strip()}` does not have a spider chart")
            continue
        spider_languages = [lang for lang in person.languages.keys()]
        spider_colors = [color_map[l] for l in spider_languages]
        ch = SpiderChart(cx=P2_PIE_COLS[i], cy=P2_SPIDER_LINE, rx=64, max_score=8)
        ch.render(g2, spider_matrix, colors=spider_colors, id_prefix=f"p2_person{i}_spiderchart")
    # page 3
    getLogger().info("Generating page 3 SVG")
    t3 = Canvas(PAGE3_TEMPLATE)
    g3 = t3.group('leo')
    # List languages
    _languages = profile.languages if profile.languages else profile.get_extra_language_legends()
    colors = [color_map[l] for l in _languages]
    for idx, (lang, color) in enumerate(zip(_languages, colors)):
        g3.text(lang.upper(), center=(P3_LANG_TEXT_COLS[idx], P3_LANG_ROW2), style=P3_LANG_NAME_STYLE, id_prefix=f"language_legends_text{idx}")
        g3.circle((P3_LANG_BALL_COLS[idx], P3_LANG_ROW2), 10, style=STYLE_REDDOT.clone(fill=color, stroke="none"), id_prefix=f"language_legends_circle_{idx}")
    # ----------------------------------------
    # draw charts & names
    for i, person in enumerate(profile.people[4:8]):
        languages = [lang for lang, p in person.languages.items() if p.dist > 0]
        colors = [color_map[l] for l in languages]
        g3.text(person.name.strip().upper(), center=(P3_PIE_COLS[i], P3_NAME_ROW), style=P3_PEOPLE_NAME_STYLE, id_prefix=f"p3_person{i}_name")
        if person.usage != -1:
            g3.text(f"{round(person.usage)}%", center=(P3_PIE_COLS[i], P3_USAGE_ROW), style=P3_PEOPLE_NAME_STYLE, id_prefix=f"p3_person{i}_usage")
        else:
            g3.text(f"- N/A -", center=(P3_PIE_COLS[i], P3_USAGE_ROW), style=P3_PEOPLE_NAME_STYLE, id_prefix=f"p3_person{i}_usage")
        # render pie charts
        slides = [x.dist for x in person.languages.values() if x.dist > 0]
        if not person.piesize:
            _adjust = 100
        elif person.piesize > 10:
            _adjust = person.piesize
        else:
            # draw draw the pie chart too small
            _adjust = 10
        radius = (PIE_RADIUS * _adjust / 100, PIE_RADIUS * _adjust / 100)
        getLogger().debug(f"{person.name} -- adjust -> {_adjust}")
        pie = PieChart(g3, center=(P3_PIE_COLS[i], P3_PIE_ROW), radius=radius, colors=colors)
        pie.slide(*slides)
        pie.render(id_prefix=f"p3_person{i}_piechart")
        # render spider charts
        spider_matrix = [x.skills for x in person.languages.values() if x.skills.is_valid()]
        if not spider_matrix:
            getLogger().warning(f"{profile.reportID}/page 3 - person `{person.name.strip()}` does not have a spider chart")
            continue
        spider_languages = [lang for lang in person.languages.keys()]
        spider_colors = [color_map[l] for l in spider_languages]
        ch = SpiderChart(cx=P3_PIE_COLS[i], cy=P3_SPIDER_ROW, rx=64, max_score=8)
        ch.render(g3, spider_matrix, colors=spider_colors, id_prefix=f"p3_person{i}_spiderchart")
    # Render SVG files
    pdf_filename_stem = profile.output_filename()[:-4] if profile.output_filename().endswith('.pdf') else profile.output_filename()
    getLogger().info("Exporting SVG files ...")
    t1.render(output_dir / f'{pdf_filename_stem}_leo_page1.svg', overwrite=overwrite)
    t2.render(output_dir / f'{pdf_filename_stem}_leo_page2.svg', overwrite=overwrite)
    t3.render(output_dir / f'{pdf_filename_stem}_leo_page3.svg', overwrite=overwrite)
    getLogger().debug("Exporting PDF pages ...")
    svg_to_pdf(output_dir / f'{pdf_filename_stem}_leo_page1.svg', overwrite=overwrite)
    svg_to_pdf(output_dir / f'{pdf_filename_stem}_leo_page2.svg', overwrite=overwrite)
    svg_to_pdf(output_dir / f'{pdf_filename_stem}_leo_page3.svg', overwrite=overwrite)
    getLogger().debug("Combining PDF pages ...")
    output_filepath = output_dir / profile.output_filename()
    if not overwrite and output_filepath.exists():
        getLogger().warning(f"{output_filepath} report file exists. SKIPPED")
    else:
        merge_pdf(output_path=output_filepath,
                  input_paths=[output_dir / f"{pdf_filename_stem}_leo_page1.pdf",
                               output_dir / f"{pdf_filename_stem}_leo_page2.pdf",
                               output_dir / f"{pdf_filename_stem}_leo_page3.pdf"])
    # generated output filename
    if auto_clean and output_filepath.is_file():
        (output_dir / f'{pdf_filename_stem}_leo_page1.svg').unlink()
        (output_dir / f'{pdf_filename_stem}_leo_page2.svg').unlink()
        (output_dir / f'{pdf_filename_stem}_leo_page3.svg').unlink()
        (output_dir / f'{pdf_filename_stem}_leo_page1.pdf').unlink()
        (output_dir / f'{pdf_filename_stem}_leo_page2.pdf').unlink()
        (output_dir / f'{pdf_filename_stem}_leo_page3.pdf').unlink()
    return output_filepath



import logging
import itertools
from pyinkscape import Point, Style
from pyinkscape.styles import BLIND_COLORS


def getLogger():
    return logging.getLogger(__name__)


class Line:
    def __init__(self, pointA, pointB):
        if not pointA or not pointB or pointA == pointB:
            raise Exception(f"Invalid line points (p1={pointA}, p2={pointB})")
        self.pointA = pointA
        self.pointB = pointB

    def sig(self):
        ''' Get this line's signature in slope/intercept form '''
        if self.pointA.x == self.pointB.x:
            # vertical line
            return (None, float(self.pointA.x))
        elif self.pointA.y == self.pointB.y:
            # horizontal line
            return (0.0, float(self.pointA.y))
        else:
            slope = (self.pointA.y - self.pointB.y) / (self.pointA.x - self.pointB.x)
            intercept = self.pointB.y - slope * self.pointB.x
            return (float(slope), float(intercept))


class Quadrangle:
    def __init__(self, top, right, bottom, left):
        """

        """
        self.top = top
        self.right = right
        self.bottom = bottom
        self.left = left

    def lines(self):
        lines = []
        if self.top != self.right:
            lines.append(Line(self.top, self.right))
        if self.right != self.bottom:
            lines.append(Line(self.right, self.bottom))
        if self.bottom != self.left:
            lines.append(Line(self.bottom, self.left))
        if self.left != self.top:
            lines.append(Line(self.left, self.top))
        return lines

    def overlap(self, lines):
        for line in self.sigs():
            if line in lines:
                return True
        return False

    def sigs(self):
        ''' All signatures (slope-intercept) of lines in this quadrangle '''
        return [line.sig() for line in self.lines()]

    def __repr__(self):
        return f"Top:{self.top}-Right:{self.right}-Bottom:{self.bottom}-Left:{self.left}"

    def __add__(self, other):
        if isinstance(other, Quadrangle):
            return Quadrangle(self.top + other.top, self.right + other.right, self.bottom + other.bottom, self.left + other.left)
        else:
            return Quadrangle(self.top + other, self.right + other, self.bottom + other, self.left + other)


class SpiderChart:

    DEFAULT_LINESTYLE = Style(fill="none", stroke="#000000", stroke_miterlimit='10')

    def __init__(self, cx=50, cy=80, rx=67, ry=None, max_score=10, flex_size=2):
        self.cx = cx  # center
        self.cy = cy
        self.rx = rx  # radius x-axis
        self.ry = ry if ry else self.rx  # radius y-axis
        self.max_score = max_score
        self.flex_size = flex_size
        self._lines = set()
        self.reset_lines()

    def reset_lines(self):
        self._lines.clear()
        self._lines.add((None, self.cx))
        self._lines.add((0, self.cy))

    def graph(self):
        ''' Draw coordinate graph (vertical & horizontal) '''
        return f"M {self.cx},{self.cy-self.ry} v {self.ry * 2} M {self.cx - self.rx},{self.cy} h {self.rx * 2}"

    def line(self, c_top, c_right, c_bottom, c_left):
        ''' Create a single category path (4 points) '''
        return f"M {c_top.x},{c_top.y} L {c_right.x},{c_right.y} L {c_bottom.x},{c_bottom.y} L {c_left.x},{c_left.y} Z"

    def draw_line(self, group, top, right, bottom, left, style=None, color=None, **kwargs):
        line = self.line(top, right, bottom, left)
        _style = style if style else SpiderChart.DEFAULT_LINESTYLE.clone(stroke_width='1px')
        if color:
            _style = _style.clone(stroke=color)
        group.path(line, style=_style, **kwargs)

    def to_quadrangle(self, top, right, bottom, left):
        ''' Convert a skill matrix (i.e. scores) to drawing coordinates '''
        top_loc = Point(self.cx, self.cy - self.ry * top / self.max_score)
        right_loc = Point(self.cx + self.rx * right / self.max_score, self.cy)
        bottom_loc = Point(self.cx, self.cy + self.ry * bottom / self.max_score)
        left_loc = Point(self.cx - self.rx * left / self.max_score, self.cy)
        return Quadrangle(top_loc, right_loc, bottom_loc, left_loc)

    def shift_quads(self, quads):
        _lines = self._lines  # don't shift if not needed
        new_quads = []
        for quad in quads:
            trial = 1  # circle
            found = None if quad.overlap(_lines) else quad
            while not found:
                delta = self.flex_size * trial
                deltas = (Point(0, delta), Point(0, -delta), Point(delta, 0), Point(-delta, 0),
                          Point(delta, delta), Point(delta, -delta), Point(-delta, delta), Point(-delta, -delta))
                for d in deltas:
                    getLogger().debug(f"Trying {d}")
                    new_quad = quad + d
                    if not new_quad.overlap(_lines):
                        found = new_quad
                        break
                if not found:
                    trial += 1
            getLogger().debug(f"  shifted -> {found}")
            new_quads.append(found)
            # remember these added lines
            for line in found.lines():
                _lines.add(line.sig())
        return new_quads, _lines  # update _lines into self._lines if new_quads are drawn

    def render(self, group, scores, colors=BLIND_COLORS, id_prefix="spiderchart", **kwargs):
        # draw graph line
        group.path(self.graph(), style=SpiderChart.DEFAULT_LINESTYLE, id_prefix=f"{id_prefix}_coordinate_graph")
        self.reset_lines()
        # draw all lines
        quads = [self.to_quadrangle(top, right, bottom, left) for top, right, bottom, left in scores]
        quads, _lines = self.shift_quads(quads)
        self._lines.update(_lines)
        for idx, (quad, color) in enumerate(zip(quads, itertools.cycle(colors)), start=1):
            self.draw_line(group, quad.top, quad.right, quad.bottom, quad.left, color=color, id_prefix=f"{id_prefix}_line{idx}")



# -*- coding: utf-8 -*-

# chirptext's package version information
__author__ = "Le Tuan Anh"
__email__ = "tuananh.le@ntu.edu.sg"
__copyright__ = "Copyright (c) 2020, BLIP lab, NTU, Singapore"
__credits__ = []
__license__ = "GNU General Public License v3 (GPLv3)"
__description__ = "Language Experience Overview (LEO) Report Generator"
__url__ = "https://blipntu.github.io/LEO/"
__maintainer__ = "Le Tuan Anh"
__version_major__ = "0.1"
__version__ = "{}a5.post1".format(__version_major__)
__version_long__ = "{} - Alpha".format(__version_major__)
__status__ = "3 - Alpha"



from .__version__ import __author__, __email__, __copyright__, __maintainer__
from .__version__ import __credits__, __license__, __description__, __url__
from .__version__ import __version_major__, __version_long__, __version__, __status__

from .models import LEOProfile, read_json
from .leo import generate_leo
from .colors import COLORS, FIXED_LANGUAGES, to_color_list

__all__ = ['LEOProfile', 'read_json', 'generate_leo', 'COLORS', 'FIXED_LANGUAGES', 'to_color_list']
